use reqwest::header::HeaderValue;
use std::time::{SystemTime, UNIX_EPOCH};
use tokio::time::{sleep, Duration};

pub struct GithubEndpointWatcher {
    username: String,
    token: String,
    url: String,
    http_client: reqwest::Client,
    etag: Option<HeaderValue>,
    next_poll_allowed: u64,
}

impl GithubEndpointWatcher {
    pub fn new(
        username: String,
        token: String,
        url: String,
    ) -> Result<Self, Box<dyn std::error::Error>> {
        let http_client = reqwest::Client::new();
        Ok(GithubEndpointWatcher {
            username,
            token,
            url,
            http_client,
            etag: None,
            next_poll_allowed: 0,
        })
    }

    pub async fn get_results(&mut self) -> Result<serde_json::Value, Box<dyn std::error::Error>> {
        let mut request = self
            .http_client
            .get(&self.url)
            .basic_auth(self.username.trim(), Some(self.token.trim()))
            .header("User-Agent", "github_watcher")
            .header("Accept", "application/vnd.github.v3+json");
        if let Some(etag) = &self.etag {
            request = request.header(reqwest::header::ETAG, etag);
        }
        self.sleep_until_next_poll().await?;
        let request_started_at = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs();
        println!("Hitting url {}", self.url);
        let response = request.send().await?;

        let headers = response.headers();
        let etag = headers.get(reqwest::header::ETAG).map(|x| x.to_owned());
        let poll_interval = headers.get("x-poll-interval").map(|x| x.to_owned());
        // let ratelimit_limit = headers.get("x-ratelimit-limit").map(|x| x.to_owned());
        // let ratelimit_remaining = headers.get("x-ratelimit-remaining").map(|x| x.to_owned());
        // let ratelimit_reset = headers.get("x-ratelimit-reset").map(|x| x.to_owned());
        // let ratelimit_used = headers.get("x-ratelimit-used").map(|x| x.to_owned());
        // let ratelimit_resource = headers.get("x-ratelimit-resource").map(|x| x.to_owned());

        let body = response.json::<serde_json::Value>().await?;
        println!("{}", serde_json::to_string(&body)?);
        self.etag = etag;
        let poll_interval_parsed: u64 = match poll_interval {
            Some(header) => header.to_str()?.parse()?,
            None => {
                println!("No poll interval returned, defaulting to 5 minute polling.");
                300
            }
        };
        self.next_poll_allowed = request_started_at + poll_interval_parsed;
        Ok(body)
    }

    async fn sleep_until_next_poll(&self) -> Result<(), Box<dyn std::error::Error>> {
        loop {
            let now = SystemTime::now().duration_since(UNIX_EPOCH)?.as_secs();
            if now >= self.next_poll_allowed {
                break;
            }
            sleep(Duration::from_secs(self.next_poll_allowed - now)).await;
        }
        Ok(())
    }
}
